/*!
	@brief	MIDI parsing library

	@file	midi.c
	@author	Satoshi "Chuck" Takagi
	@date	2006/12/09 09:03:57
 */
/* 
	$RCSfile: midi.c,v $
	$Revision: 143 $
	$Date: 2008-12-30 06:22:53 +0900 (火, 30 12 2008) $
 */

#include <inttypes.h>
#include <avr/io.h>
#include <avr/eeprom.h>
#include "global.h"
#include "midi.h"


extern uint8_t rb;				//!< read byte from UART
extern uint8_t g_midi_ch;		//!< MIDI channel
static uint8_t midi_stat = 255;	//!< MIDI parsing status initial value
static uint8_t midi_recv_ch;	//!< MIDI receive channel
static uint8_t midi_byte[3];	//!< MIDI parsing work array

/*!
	@brief	MIDI System Message parser

	@param	なし
	@return	引き続くMIDI Channel Messageがあれば1

	@note	MIDIシステムメッセージを解析し、system exclusiveは除外してそれ以外の必要なメッセージをMidiSysMes構造体に記録する。
*/
uint8_t MidiSystemMessage(void)
{
	static uint8_t sysex;

#ifndef NO_USE
	if (sysex) {					// in sysex process
		if (rb == 0xF7) sysex=0;	// end of sysex process
	} else {
		if (rb < 0xF8) { 
			if (rb > 0x7F) {		// voice msg
				if (rb == 0xF0) 	sysex = 1;		// start of sysex process
				else 				midi_recv_ch = rb & 0x0F;
				midi_stat = 0;
			} else {				// store midi data byte 00-7F
				midi_byte[midi_stat & 0x01] = rb;
 			}
 			return 1;
		}
	}
#else
	if (sysex) {					// in sysex process
		if (rb == 0xF7) sysex=0;	// end of sysex process
	} else {
		if (rb < 0xF8) { 
			if (rb >= 0xF0) {		// common msg
				if (rb == 0xF0)	sysex = 1;	// start of sysex process
				midi_stat = 20;
			} else if (rb > 0x7F) {	// voice msg
				midi_recv_ch = rb & 0x0F;
				if (midi_recv_ch == g_midi_ch) {
					midi_stat = 0;
				}
			} else {				// store midi data byte 00-7F
				midi_byte[midi_stat & 0x01] = rb;
 			}
 			return 1;
 		} else {					// realtime msg >= 0xF8
			switch (rb) {
			case 0xF8:	break;		// timing clock
			case 0xF9:	break;
			case 0xFA:	break;		// start
			case 0xFB:	break;		// continue
			case 0xFC:	break;		// stop
			case 0xFD:	break;
			case 0xFE:	MidiSysMes.Status |= fMidiActiveSensing; break;		// active sensing
			case 0xFF:	MidiSysMes.Status |= fMidiSystemReset;   break;		// system reset
			}
		}
	}
#endif
	return 0;
}

/*!
	@brief	MIDI Channel Message parser

	@param	なし
	@return	なし

	@note	MIDIチャネルメッセージを解析し、必要なメッセージをMidiChanMes構造体に記録する。1回のコールにつき1バイトずつ処理し、進捗状態等をmidi_byteにて管理している。
*/
void MidiChannelMessage(void)
{
	switch (midi_stat) {
	case 0:
		switch ( rb & 0xF0 ) {
#ifdef MCV2313A
		case MIDI_NOTE_OFF : midi_byte[2] = rb; midi_stat=2;  break;
		case MIDI_NOTE_ON  : midi_byte[2] = rb; midi_stat=4;  break;
		case MIDI_PBEND    : midi_byte[2] = rb; midi_stat=14; break;
#endif // MCV2313A
		case MIDI_C_CHG    : midi_byte[2] = rb; midi_stat=10; break;
#ifdef NO_USE
		case MIDI_P_PLS    : midi_byte[2] = rb; midi_stat=6;  break;
		case MIDI_P_CHG    : midi_byte[2] = rb; midi_stat=8;  break;
		case MIDI_C_PLS    : midi_byte[2] = rb; midi_stat=12; break;
#endif
		default:
			midi_stat=255;
		}
		break;

#ifdef MCV2313A
	/* Note Off */
	case 2: midi_stat=3; break;
	case 3: midi_stat=2;
			if (midi_recv_ch == g_midi_ch) {
				MidiChanMes.NoteOff.Note = midi_byte[0];
				MidiChanMes.NoteOff.Velocity = midi_byte[1];
				MidiChanMes.Status |= fMidiNoteOff;
			}
			break;

	/* Note On */
	case 4: midi_stat=5; break;
	case 5: midi_stat=4;
			if (midi_recv_ch == g_midi_ch) {
				if (midi_byte[1] == 0) {
					// note off
					MidiChanMes.NoteOff.Note = midi_byte[0];
					MidiChanMes.NoteOff.Velocity = 64;
					MidiChanMes.Status |= fMidiNoteOff;
				} else {
					MidiChanMes.NoteOn.Note = midi_byte[0];
					MidiChanMes.NoteOn.Velocity = midi_byte[1];
					MidiChanMes.Status |= fMidiNoteOn;
				}
			}
			break;

	/* PitchBend */
	case 14: midi_stat=15; break;
	case 15: midi_stat=14;	//
			if (midi_recv_ch == g_midi_ch) {
				MidiChanMes.PitchBend.Value = (midi_byte[1]<<7) + midi_byte[0];
				MidiChanMes.Status |= fMidiPitchBend;
			}
			break;
#endif // MCV2313A

	/* Control Change */
	case 10: midi_stat=11; break;
	case 11: midi_stat=10; // for running status
			if (midi_recv_ch == g_midi_ch) {
				MidiChanMes.ControlChange.Control = midi_byte[0];
				MidiChanMes.ControlChange.Value = midi_byte[1];
				if (midi_byte[0] < 120)
					MidiChanMes.Status |= fMidiControlChange;
				else
					MidiChanMes.Status |= fMidiModeChange;
			}
			break;

#ifdef NO_USE
	/* Polyphonic Key Pressure */ 
	case 6: midi_stat=7; break;
	case 7: //midi_stat=6;	//
				MidiChanMes.PolyPressure.Note = midi_byte[0];
				MidiChanMes.PolyPressure.Value = midi_byte[1];
				MidiChanMes.Status |= fMidiPolyPressure;
			break;

	/* Program change */
	case 8:
				MidiChanMes.ProgramChange.Value = midi_byte[0];
				MidiChanMes.Status |= fMidiProgramChange;
			break;

	/* ChannelPressure */
	case 12: 
				MidiChanMes.ChannelPressure.Value = midi_byte[0];
				MidiChanMes.Status |= fMidiChannelPressure;
			break;

	case 20:
		switch ( rb ) {
		case MIDI_MIDITC   : midi_stat =  22; break;
		case MIDI_SONG_PTR : midi_stat =  24; break;
		case MIDI_SONG_SEL : midi_stat =  26; break;
		case MIDI_TUNE_REQ : midi_stat = 255; break;
		case 0xF4          : midi_stat = 255; break;
		case 0xF5          : midi_stat = 255; break;
		}
		break;

	/* MIDI timecode */
	case 22: break;

	/* Song Position Pointer */
	case 24: midi_stat=25; break;
	case 25: break;

	/* Song Select */
	case 26: break;
#endif
	
	default: break;
	}

}
